/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*! \file Attribute.h
  \brief Contains the Attribute class.
  
  Attribute provides a non-templated class to interface with over the templated ba and partvector classes.
*/
/*
  TO DO LIST FOR EV. PLEASE DO NOT CHANGE ANYTHING!
2) operations on fftw_complex type are not defined yet.
3) cross product for vector2 and vector3 to be added (defined in Vector2/3 but not in Attribute)
4) Some operations on Vector2 and Vector3 are to be defined and added.
5) serialization functions.
6) Some collective operations
*/
#ifndef ATTRIBUTE_H
#define ATTRIBUTE_H
const double BA_ADD_RESERVE=0;
#include "ba_trait.h"
#include "BA.h"
#include "BA.C"
#include "Context.h" //QQ should we include Context.h here?
class Data;
class List;
class Grid;


typedef  boost::variant< ba<double>, ba<float>, ba<int>, ba<Vector3>, ba<char>,ba<Vector2>, ba<fftw_complex> > bv;
typedef  boost::variant<double,float,int,fftw_complex> bv_pod;
//typedef  boost::variant<fftw_complex> bv_complex;
typedef  boost::variant<vector<aux_vect<double> >,vector<aux_vect<float> >,vector<aux_vect<int> > > bv_stat;
typedef  boost::variant<Vector3,Vector2 > bv_vec;
typedef  boost::variant< double, int, float, Vector3, Vector2, char, fftw_complex > bv1;
typedef  boost::variant<int> bv2;
typedef  boost::variant<unsigned int> bv3;
typedef  boost::variant<vector<int> > bv_stlvec;
//============================================================================
enum Attribute_Data_Type {BASIN_DATA_NONE,BASIN_DATA_INT,BASIN_DATA_FLOAT,BASIN_DATA_DOUBLE,BASIN_DATA_CHAR,BASIN_DATA_COMPLEX,BASIN_DATA_VECTOR2,BASIN_DATA_VECTOR3};


// Auxiliary global function
template <typename T> Attribute_Data_Type find_type(T val)
{
  Attribute_Data_Type temp;
  if (typeid(val)==typeid(int))
    {
      temp=BASIN_DATA_INT;
    }
  else if (typeid(val)==typeid(float))
    {
      temp=BASIN_DATA_FLOAT;
    }
  else if (typeid(val)==typeid(double))
    {
      temp=BASIN_DATA_DOUBLE;
    }
  else if (typeid(val)==typeid(char))
    {
      temp=BASIN_DATA_CHAR;
    }
  else if (typeid(val)==typeid(fftw_complex))
    {
      temp=BASIN_DATA_COMPLEX;
    }
  else if (typeid(val)==typeid(Vector2))
    {
      temp=BASIN_DATA_VECTOR2;
    }
  else if (typeid(val)==typeid(Vector3))
    {
      temp=BASIN_DATA_VECTOR3;
    }
  else
    {
      temp=BASIN_DATA_NONE;
    }
  return temp;
};

inline Attribute_Data_Type find_variant_type(const  bv& ta) 
{
  Attribute_Data_Type temp;
  if (ta.type()==typeid(ba<int>))
    {
      temp=BASIN_DATA_INT;
    }
  else if (ta.type()==typeid(ba<float>))
    {
      temp=BASIN_DATA_FLOAT;
    }
  else if (ta.type()==typeid(ba<double>))
    {
      temp=BASIN_DATA_DOUBLE;
    }
  else if (ta.type()==typeid(ba<char>))
    {
      temp=BASIN_DATA_CHAR;
    }
  else if (ta.type()==typeid(ba<fftw_complex>))
    {
      temp=BASIN_DATA_COMPLEX;
    }
  else if (ta.type()==typeid(ba<Vector2>))
    {
      temp=BASIN_DATA_VECTOR2;
    }
  else if (ta.type()==typeid(ba<Vector3>))
    {
      temp=BASIN_DATA_VECTOR3;
    }
  else
    {
      temp=BASIN_DATA_NONE;
    }
  return temp;
};




//===================START OF BV1======================================

/* BV1 is an auxiliary class which allows to automatically cast output of 
   the [] operator. This will be put in a separate file later.
*/
class BV1
{
 public:
  bv1 value;
  BV1(bv1 initial){value=initial;}
  template<typename T> operator T() const{
    try
      {
	T to_Return;
	to_Return=boost::get<T>(value);
	return to_Return;	
      }
    catch(boost::bad_get &bg)
      {
	cerr<<"Error:the values in the attribute are of a different type"<<endl;
      }
  };


  
};
/* This is out because of template specialization not allowed in class*/

  template<> inline BV1::operator double() const{
    try
      {
	double to_Return;
	to_Return=boost::get<double>(value);
	return to_Return;	
      }
    catch(boost::bad_get &bg)
      {
	try
	  {
	    double to_Return;
	    to_Return=boost::get<float>(value);
	    return to_Return;	
	  }
	catch(boost::bad_get &bg)
	  {
	    double to_Return;
	    to_Return=boost::get<int>(value);
	    return to_Return;	
	  }
      }
  }

  template<> inline BV1::operator float() const{
    try
      {
	float to_Return;
	to_Return=boost::get<float>(value);
	return to_Return;	
      }
    catch(boost::bad_get &bg)
      {
	try
	  {
	    float to_Return;
	    to_Return=boost::get<double>(value);
	    return to_Return;	
	  }
	catch(boost::bad_get &bg)
	  {
	    float to_Return;
	    to_Return=boost::get<int>(value);
	    return to_Return;	
	  }
      }
  }
 
//====================END OF BV1====================================

#include "visitors.C"

/** \class Attribute Attribute.h "Attribute.h"
    \brief The Attribute class is a general container for storing data.

    A Attribute is based off a \link ba BasinArray\endlink,
    and thus are distributed objects.  
    In general, users should deal with this container only through the context
    of a List or Grid.  The purpose of Attribute is to provide a non-templated
    interface over the templated ba and partvector
*/

class Attribute{
 public:

  /** @name Attribute Constructors
      Functions to create Attributes, must be called by all nodes.
      
      Constructors must be passed a size either through an integer length (1d)
      or a vector<int> containing the size in each dimension (Nd).
      The second argument is a dummy value to set the type information.
  */
  //@{
  
  //! Empty Attribute constructor
  Attribute() {
    data=NULL;
    attribute_context=new Context; 
    attribute_name="";
    data_type=BASIN_DATA_NONE;
  }
  /// Attribute copy
  Attribute(const Attribute& y) {
    data=NULL;
    attribute_context=new Context;
    attribute_context=y.get_context_pointer();
    attribute_name=y.get_name();
    data_type=y.get_data_type();
    bv temp;
    y.get_data(temp);
    this->set_attribute(temp);
  }

  Attribute(const bv &ta)
    {
      data=new bv;
      *data=ta;
      attribute_context=new Context;
      attribute_name="";
      data_type=find_variant_type(ta);//WILL IT WORK?SET TYPE HERE
    }



  /// Attribute constructor using vector<int>
  template<class T> Attribute(vector<int> n,T newData)
    {
      attribute_name="";
      int ndims=n.size();
      vector<double> ll(ndims,0.),ur(ndims,1.);
      attribute_context=new Context(ll,ur,n,NULL);
      data_type=find_type(newData);
      data=NULL;
      ba<T> temp;
      temp.ba_create(n,BASIN_COMM,0);
      temp.ba_fill(newData);
      set_attribute(temp);
  }
  
  /// The specific 1d version of the constructor
  template<class T> Attribute(int nIndex,T newData)
    {
      //      attribute_context=NULL;
      attribute_name="";
      data_type=find_type(newData);
      data=NULL;
      ba<T> temp;
      vector<int> n;
      n.push_back(nIndex);
      temp.ba_create(n,BASIN_COMM,0);
      temp.ba_fill(newData);

      int ndims=n.size();
      vector<double> ll(ndims,0.),ur(ndims,1.);
      attribute_context=new Context(ll,ur,n,NULL);

      set_attribute(temp);
  }


  template<class T> Attribute(vector<int> n,T newData,string name,Context *c)
    {
      attribute_context=new Context;
      attribute_context=c;
      attribute_name=name;
      data_type=find_type(newData);
      data=NULL;
      ba<T> temp;
      temp.ba_create(n,BASIN_COMM,0);
      temp.ba_fill(newData);
      set_attribute(temp);
  }
  
  /// The specific 1d version of the constructor
  template<class T> Attribute(int nIndex,T newData,string name,Context *c)
    {
      attribute_context=new Context;
      attribute_context=c;
      attribute_name=name;
      data_type=find_type(newData);
      data=NULL;
      ba<T> temp;
      vector<int> n;
      n.push_back(nIndex);
      temp.ba_create(n,BASIN_COMM,0);
      temp.ba_fill(newData);
      set_attribute(temp);
  }


  Attribute(const bv &ta,Context *c)
    {
      data=new bv;
      *data=ta;
      attribute_context=new Context;
      attribute_context=c;
      attribute_name="";
      data_type=find_variant_type(ta);
    }



  //@}
  /** @name Attribute Destructor
      Destructor for Attribute... currently leaks memory.*/
  //@{
  /// Destructor to clear memory
  ~Attribute() 
    {
      int FLAG;
      MPI_Finalized(&FLAG);
      /*
	for attributes going out of scope at the end of a program
	after mpi_finalize has been called the destructor is not called
	as the mpi functions used in the destructor can not be used any more
      */
      if (!FLAG)
	{
	  if (data!=NULL) delete this->data;
	  this->data=NULL;
	}

    } 
  //@}
  
  Attribute& operator=(const Attribute &b)
    {
      if(this==&b)
	{
	  cerr<<"what am i doing?"<<endl;
	  return *this;
	}
      
      bv temp;
      b.get_data(temp);
      this->set_attribute(temp);
      this->set_context(b.get_context_pointer());
      data_type=b.get_data_type();
      return *this;
    }



  /** @name Attribute Accessor Functions
      Functions intended to retrieve data for calculations.*/
  //@{

  /// Returns the entire Attribute by argument




  void set_attribute(const bv &ta)
    {
      if (this->data==NULL) data=new bv;
      *data=ta;
      data_type=find_variant_type(ta);//WILL IT WORK?SET TYPE HERE
      //      data_type=SET TYPE HERE
    }

  
  void get_data(bv &ta) const {
    ta=*(data);
  }

  Attribute_Data_Type get_data_type() const
    {
      return data_type;
    }
  
  template<class T> int get_data_elements(T* &outVal) const {


    if (ba<T> *loc=(boost::get<ba<T> >( this->data))){


    vector<int> lo;
    for(int i=0;i<int((loc->ba_get_dims()).size());i++)
      lo.push_back(0);

    vector<int> hi = loc->ba_get_dims();

    T *temp=new T[loc->ba_size()];
    loc->ba_get(lo,hi,temp);
    outVal=&(temp[0]);
    }
    return 0;

  }

  vector<bv1> get_data_elements(int iLo, int iHi) const 
    {
      vector<int> index;
      index.push_back(iLo);
      index.push_back(iHi);
      bv_stlvec loc_index=index;
      return boost::apply_visitor(Attribute_data_to_vec_visitor(),*data,loc_index);
    }



  /// Return (via the second argument) the value of an element in a Attribute.
  template<typename T> int get_element(vector<int> index,T &outVal) 
    {
      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
    {
      (boost::get<ba<T> >( this->data)->ba_get(index,index,&outVal));
    }
  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
  return 1;
    }
  /// 1d version
  template<class T> int get_element(int iIndex, T &outVal) 
    {
      vector<int> index;
      index.push_back(iIndex);
      
      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_get(index,index,&outVal));
	}
      else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
      return 1;
    }

  /// Return a specific range of the Attribute
  template<class T>int get(vector<int> iLo, vector<int> iHi, T* outVal)
    {
      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
    {
      (boost::get<ba<T> >( this->data)->ba_get(iLo,iHi,outVal));
    }
  else cerr<<"WARNING: attribute not returned. Type mismatch"<<endl;
  return 1;

    }

  /// 1d version
  template<class T>int get(int iLo, int iHi, T* outVal)
    {
      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_get(iLo,iHi,outVal));
	}
      else cerr<<"WARNING: attribute not returned. Type mismatch"<<endl;
      return 1;
    }


  /** @name Attribute Modifier Functions
      Functions to alter or set the data in a Attribute.*/
  //@{
  /// Replace an element of a Attribute
  template<class T> int set_element(  vector<int> pos, T &newData) {
   if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
    {
      (boost::get<ba<T> >( this->data)->ba_putF(pos,pos,&newData));
    }
  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
  return 1;

  }
  /// 1d version
  template<class T> int set_element(  int iPos, T &newData) {
    vector<int> pos;
    pos.push_back(iPos);
    if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
      {
	(boost::get<ba<T> >( this->data)->ba_putF(pos,pos,&newData));
      }
    else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
    return 1;
  }

  /// Add a value to the end of a Attribute
  /*  template<class T> int append( T& newData) {
    return data->append(newData);
    }*/

  /// Put data into a specific range of the Attribute
  template<class T>int put(vector<int> iLo,vector<int> iHi, T* newData)
    {

      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_putF(iLo,iHi,newData));
	}
      else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
      return 1;
      
    }
  /// 1d version
  template<class T>int put(int iLo,int iHi, T* newData)
    {
      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_put(iLo,iHi,newData));
	}
      else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
      return 0;
      
    }  

  //These are new versions of "put" which use stl vectors. 
  //While I am going to keep the old version of functions using arrays
  //I am going to create new versions using stl containers. EV
  template<class T>int put(vector<int> iLo,vector<int> iHi, vector<T> &newData)
    {

      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_putF(iLo,iHi,&newData[0]));
	}
      else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
      return 1;
      
    }


  template<class T>int put(int iLo,int iHi, vector<T> &newData)
    {
      if (newData.size()>=(iHi-iLo+1))
	{
	  if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	    {
	      (boost::get<ba<T> >( this->data)->ba_put(iLo,iHi,&newData[0]));
	    }
	  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
	}
      else cerr<<"WARNING: attribute not set. Not enough data in the vector"<<endl;

      return 0;
      
    }  

  /// Add values to elements of a BA specified by the indexes
  template<class T>int add(vector<int> iLo,vector<int> iHi, vector<T> &newData)
    {

      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_add(iLo,iHi,&newData[0]));
	}
      else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
      return 1;
      
    }


  template<class T>int add(int iLo,int iHi, vector<T> &newData)
    {
      if (newData.size()>=(iHi-iLo+1))
	{
	  if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	    {
	      (boost::get<ba<T> >( this->data)->ba_add(iLo,iHi,&newData[0]));
	    }
	  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
	}
      else cerr<<"WARNING: attribute not set. Not enough data in the vector"<<endl;

      return 0;
      
    }  
  //===========NEW
  /// Multiply values to elements of a BA specified by the indexes
  template<class T>int multiply(vector<int> iLo,vector<int> iHi, vector<T> &newData)
    {

      if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	{
	  (boost::get<ba<T> >( this->data)->ba_multiply(iLo,iHi,&newData[0]));
	}
      else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
      return 1;
      
    }


  template<class T>int multiply(int iLo,int iHi, vector<T> &newData)
    {
      if (newData.size()>=(iHi-iLo+1))
	{
	  if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
	    {
	      (boost::get<ba<T> >( this->data)->ba_multiply(iLo,iHi,&newData[0]));
	    }
	  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
	}
      else cerr<<"WARNING: attribute not set. Not enough data in the vector"<<endl;

      return 0;
      
    }  






  //========NEW
  /// Change a specific element
  template<class T>int amend(vector<int> pos, T newData)
    {
  if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
    {
      (boost::get<ba<T> >( this->data)->ba_putF(pos,pos,&newData));
    }
  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
  return 0;

    }
  
  /// 1d version
  template<class T>int amend(int iPos, T newData)
    {
      vector<int> pos;
      pos.push_back(iPos);
 if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
    {
      (boost::get<ba<T> >( this->data)->ba_put(pos,pos,&newData));
    }
  else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
  return 0;

    }
  
  /// Fill the Attribute with one value
  template<class T> int fill(T tConst){

    if (ba<T> *loc=(boost::get<ba<T> >( this->data)))
      {
	(boost::get<ba<T> >( this->data)->ba_fill(tConst));
      }
    else cerr<<"WARNING: attribute not set. Type mismatch"<<endl;
    return 0;
  }
   //@}

  /** @name Utilities
      Functions to get useful information about the Attribute.*/
  //@{

  /// Return the size of the Attribute
  int size() const   { 

     return boost::apply_visitor(Attribute_size_visitor(),*data);
  }



  /// Gets the range of rows held by a process locally
  int get_gen_distribution(int procRank,int* iLo, int* iHi) 
    {
      bv2 loc_rank=procRank;
      vector<int> local_limits(2);
     local_limits=boost::apply_visitor(Attribute_distribution_visitor(),*data,loc_rank);
     *iLo=local_limits[0];
     *iHi=local_limits[1];
      return 0;

    }
  
 /// Gets the range of indices held by a process locally
  int get_gen_distribution_index(int procRank, int* iLo, int *iHi)
    {
      bv2 loc_rank=procRank;
      vector<int> local_limits(2);
     local_limits=boost::apply_visitor(Attribute_distributionIndex_visitor(),*data,loc_rank);
     *iLo=local_limits[0];
     *iHi=local_limits[1];

      return 0;
    }



  //=========Versions returning vector<int>===============

  /// Gets the range of rows held by a process locally
  vector<int> get_gen_distribution(int procRank) 
    {
      bv2 loc_rank=procRank;
      vector<int> local_limits(2);
     local_limits=boost::apply_visitor(Attribute_distribution_visitor(),*data,loc_rank);
      return local_limits;

    }
  
 /// Gets the range of indices held by a process locally
  vector<int> get_gen_distribution_index(int procRank)
    {
      bv2 loc_rank=procRank;
      vector<int> local_limits(2);
      local_limits=boost::apply_visitor(Attribute_distributionIndex_visitor(),*data,loc_rank);
      return local_limits;
    }

  //=======  

  /// Get the dimensions of an Attribute.
  /// Example: for a 2-d Attribute of 10x30 elements this functions returns
  /// a vector<int> with elements (10,30)
  vector<int> get_dims() const 
    {
      if (data==NULL)
	{ 
	  vector<int> to_Return(1,0);
	  return to_Return;
	}
      else
	{
	  return boost::apply_visitor(Attribute_dims_visitor(),*data);
	}
    }

  //@}





  ///make the Attribute sortable
  template <typename T> int g_make_sortable(T *y, int lo, int hi, bv_stat &aux)
    {
      /* y, lo, and hi are retained only for compatibility reasons
	 but y, lo and hi will always be those of the local values
	 of each process. So we don't need to pass them
      */
      //      aux=boost::apply_visitor(Attribute_makeSortable_visitor(),*data);
      cerr<<"No longer needed"<<endl;
      return 0;
    }
  template <typename T> int g_make_sortable(bv_stat &aux)
    {
      /*
	This function is called inside a_sort
      */
      //      aux=boost::apply_visitor(Attribute_makeSortable_visitor(),*data);
      cerr<<"No longer needed"<<endl;
      return 0;
    }
  
  
  /// Sort function
  template<typename T> int g_sort(vector<aux_vect<T> > &aux)
    {
      /* This function is retained only for compatibility.
	 It requires g_make_sortable to be called before
      */
      //      aux=boost::apply_visitor(Attribute_sort_visitor(),*data);
      return 0;
    }
  
  template<typename T> int a_sort(vector<aux_vect<T> > &aux)
    {
      /* Here make_sortable is called inside the sort_visitopr*/
      bv_stat temp=boost::apply_visitor(Attribute_sort_visitor(),*data);
      aux=(boost::get<vector<aux_vect<T> > >(temp));
      return 0;
    }
  





  /*    bv1 operator[](int index)
    {
      bv2 loc_index=index;
     return boost::apply_visitor(Attribute_bracket_visitor(),*data,loc_index);
    }
  */

  BV1 get_element_variant(vector<int> index)
    {
      bv_stlvec loc_index=index;
      return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }

  //This is different from the following set_name function in that
  //It will not check to see if you are clobbering its name.
  //Users should stick to just set_name, however, this function
  //will be called from a parent list/grid, which will have already changed
  //its map entry to reflect the name change, making it safe
  bool set_name_from_parent (string name);

  void set_name(string name)
    {
      
      if (this->attribute_name=="")
	{
	  this->attribute_name=name;
	}
      else
	{
	  cerr<<"WARNING: Name changed but this Attribute already had a name: "<<this->attribute_name<<endl;
	  this->attribute_name=name;
	}
    }
  
  string get_name() const
    {
      return this->attribute_name;
    }

  void set_context(Context *c)
    {
      /*
	This function is to be used when a List or a Grid is adding an Attribute
	to its map of data(when a file is read). 
        In that case the Data add_attribute function is 
	inserting the Attribute in the map and we just need to set 
        the context of the Attribute before this is inserted in the map.
      */
      /*
	The line below is not necessary anymore, since we are creating the new 
	Context in the Constructor
	attribute_context=new Context;
      */
      this->attribute_context=c;
    }

  Context* get_context_pointer() const
    {
      return this->attribute_context;
    }


    BV1 operator[](int index)
    {
      bv2 loc_index=index;
     return boost::apply_visitor(Attribute_bracket2_visitor(),*data,loc_index);
    }


    BV1 operator()(int index)
    {
      bv2 loc_index=index;
     return boost::apply_visitor(Attribute_bracket2_visitor(),*data,loc_index);
    }


    BV1 operator()(int index1,int index2)
    {
      vector<int> temp(2);
      temp[0]=index1;
      temp[1]=index2;
      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }



    BV1 operator()(int index1,int index2, int index3)
    {
      vector<int> temp(3);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }


    BV1 operator()(int index1,int index2, int index3,int index4)
    {
      vector<int> temp(4);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }



    BV1 operator()(int index1,int index2, int index3,int index4,int index5)
    {
      vector<int> temp(5);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;
      temp[4]=index5;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }

    BV1 operator()(int index1,int index2, int index3,int index4,int index5,int index6)
    {
      vector<int> temp(6);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;
      temp[4]=index5;
      temp[5]=index6;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }


    BV1 operator()(int index1,int index2, int index3,int index4,int index5,int index6,int index7)
    {
      vector<int> temp(7);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;
      temp[4]=index5;
      temp[5]=index6;
      temp[6]=index7;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }


    BV1 operator()(int index1,int index2, int index3,int index4,int index5,int index6,int index7,int index8)
    {
      vector<int> temp(8);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;
      temp[4]=index5;
      temp[5]=index6;
      temp[6]=index7;
      temp[7]=index8;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }


    BV1 operator()(int index1,int index2, int index3,int index4,int index5,int index6,int index7,int index8,int index9)
    {
      vector<int> temp(9);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;
      temp[4]=index5;
      temp[5]=index6;
      temp[6]=index7;
      temp[7]=index8;
      temp[8]=index9;

      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }

    BV1 operator()(int index1,int index2, int index3,int index4,int index5,int index6,int index7,int index8,int index9,int index10)
    {
      vector<int> temp(10);
      temp[0]=index1;
      temp[1]=index2;
      temp[2]=index3;
      temp[3]=index4;
      temp[4]=index5;
      temp[5]=index6;
      temp[6]=index7;
      temp[7]=index8;
      temp[8]=index9;
      temp[9]=index10;
      bv_stlvec loc_index=temp;
     return boost::apply_visitor(Attribute_bracketmd_visitor(),*data,loc_index);
    }







    /*    bv1 operator[](const Attribute &index)
    {
      bv temp_index;
      bv2 loc_index;
      const_cast<Attribute &>(index).get_data(temp_index);
      loc_index=temp_index;
     return boost::apply_visitor(Attribute_bracket_visitor(),*data,loc_index);
    }
    */

 Attribute operator<(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_lessthan_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator<(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_lessthan_visitor(),*data,temp2));
      temp.set_context(this->get_context_pointer());
	return temp;
      }


 Attribute operator>(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_largerthan_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator>(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_largerthan_visitor(),*data,temp2));
      temp.set_context(this->get_context_pointer());
	return temp;
      }


 Attribute operator>=(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_largereq_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator>=(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_largereq_visitor(),*data,temp2));
      temp.set_context(this->get_context_pointer());
	return temp;
      }



 Attribute operator<=(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_lesseq_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator<=(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_lesseq_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }




 Attribute operator==(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_logicaleq_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator==(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_logicaleq_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }


 Attribute operator!=(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_logicaldiff_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator!=(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_logicaldiff_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }





    // OPERATORS +,-,/,*
    //!
    Attribute operator+(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      //We can use plusII since U is in this case ba<type>
      temp.set_attribute(boost::apply_visitor(Attribute_plusII_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }



    Attribute operator+(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_plusII_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }

    //!
    Attribute operator-(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      //We can use plusII since U is in this case ba<type>
      temp.set_attribute(boost::apply_visitor(Attribute_minus_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }


    
    Attribute operator-(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_minus_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }

    //!
    Attribute operator*(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      //We can use plusII since U is in this case ba<type>
      temp.set_attribute(boost::apply_visitor(Attribute_times_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }


    
    Attribute operator*(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_times_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }




    //!
    Attribute operator/(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      //We can use plusII since U is in this case ba<type>
      temp.set_attribute(boost::apply_visitor(Attribute_divide_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }


    Attribute operator/(bv_pod b)
      {
	Attribute temp;
	bv_pod temp2;
	temp2=b;
	temp.set_attribute(boost::apply_visitor(Attribute_divide_visitor(),*data,temp2));
	temp.set_context(this->get_context_pointer());
	return temp;
      }




    Attribute operator&&(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_times_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }


    Attribute operator||(const Attribute &b){
      Attribute temp;
      bv b_data;
      b.get_data(b_data);
      temp.set_attribute(boost::apply_visitor(Attribute_logicor_visitor(),*data,b_data));
      temp.set_context(this->get_context_pointer());
      return temp;
    }


  string get_client_information (string info = "", string depth = "");


 private:
  /// The basin variant actually containing our data
  bv *data;
  ///the name of the Attribute
  string attribute_name;
  //The Attribute Context
  Context *attribute_context;
  //type of data in the Attribute
  Attribute_Data_Type data_type;
};





//Functions
//!
Attribute attribute_random(vector<int> b,unsigned int seed=0);
//!
Attribute attribute_random(int b,unsigned int seed=0);

//Returns an attribute of the same size and context as a, but with 
// elements b(i,j,k)=i OR j OR k, depending on dir
Attribute& attribute_index(Attribute& a, int dir);
Attribute& attribute_index(vector<int> nx, int dir, int order_type=0);

// Takes an attribute attr and recenters it with new_ll as the
// lower-left corner
Attribute& recenter(Attribute& attr, vector<int> new_ll);

// Takes an attribute, and an attribute of integers, id, and resorts
// the output such that b[i]=a[id[i]]
Attribute& resort(Attribute& attr, Attribute& id);

Attribute where(const Attribute& a, const Attribute &h, const Attribute &j);

Attribute where(const Attribute& a, const double &h, const Attribute &j);

Attribute where(const Attribute& a, const float &h, const Attribute &j);

Attribute where(const Attribute& a, const int &h, const Attribute &j);

Attribute where(const Attribute& a, const fftw_complex &h, const Attribute &j);

Attribute where(const Attribute& a, const Vector3 &h, const Attribute &j);

Attribute where(const Attribute& a, const Vector2 &h, const Attribute &j);

Attribute where(const Attribute& a, const char &h, const Attribute &j);

Attribute where(const Attribute& a, const Attribute &j,const double &h);

Attribute where(const Attribute& a, const Attribute &j,const float &h);

Attribute where(const Attribute& a, const Attribute &j,const int &h);

Attribute where(const Attribute& a, const Attribute &j,const fftw_complex &h);

Attribute where(const Attribute& a, const Attribute &j,const Vector3 &h);

Attribute where(const Attribute& a, const Attribute &j,const Vector2 &h);

Attribute where(const Attribute& a, const Attribute &j,const char &h);

Attribute where(const Attribute& a, const int &j,const int &h);

Attribute where(const Attribute& a, const float &j,const float &h);

Attribute where(const Attribute& a, const double &j,const double &h);


// This is new.  The output will generally be shorter than h!!!!
Attribute where(const Attribute& a, Attribute &h);

//!
Attribute pow(const Attribute& a, double b);
//
Attribute pow(const Attribute& a, fftw_complex b);
//!
Attribute real(const Attribute& a);
//!
Attribute imag(const Attribute& a);
//!
Attribute log(const Attribute& a);
//!
Attribute log10(const Attribute& a);
//!
Attribute sin(const Attribute& a);
//!
Attribute cos(const Attribute& a);
//!
Attribute cosh(const Attribute& a);
//!
Attribute sinh(const Attribute& a);
//!
Attribute acos(const Attribute& a);
//!
Attribute asin(const Attribute& a);
//!
Attribute acosh(const Attribute& a);
//!
Attribute asinh(const Attribute& a);
Attribute tan(const Attribute& a);
//!
Attribute atan(const Attribute& a);
//!
Attribute abs(const Attribute& a);
//!
Attribute sqrt(const Attribute& a);
//!
Attribute exp(const Attribute& a);
//!
Attribute exp10(const Attribute& a);
//!
Attribute sort(const Attribute& a);
//!
Attribute index_sort(const Attribute& a);
// Computes the complex conjugate of a complex attribute, otherwise,
// return the input
Attribute& conj(Attribute& a);
// Transforms an attribute into integers.
Attribute& fix(const Attribute& a);
// Computes the minimum of an attribute.  Returns double by default.
double min(Attribute& a);
// Computes the maximum of an attribute.  Returns double by default.
double max(Attribute& a);
BV1 sum(const Attribute& a);

//userf's still needed
Attribute operator+(int a, const Attribute &b);
Attribute operator+(float a, const Attribute &b);
Attribute operator+(double a, const Attribute &b);
Attribute operator+(fftw_complex a, const Attribute &b);
Attribute operator+(Vector2 a, const Attribute &b);
Attribute operator+(Vector3 a, const Attribute &b);

Attribute operator-(int a, const Attribute &b);
Attribute operator-(float a, const Attribute &b);
Attribute operator-(double a, const Attribute &b);
Attribute operator-(fftw_complex a, const Attribute &b);
Attribute operator-(Vector2 a, const Attribute &b);
Attribute operator-(Vector3 a, const Attribute &b);

Attribute operator*(int a, const Attribute &b);
Attribute operator*(float a, const Attribute &b);
Attribute operator*(double a, const Attribute &b);
Attribute operator*(fftw_complex a, const Attribute &b);
Attribute operator*(Vector2 a, const Attribute &b);
Attribute operator*(Vector3 a, const Attribute &b);

Attribute operator/(int a, const Attribute &b);
Attribute operator/(float a, const Attribute &b);
Attribute operator/(double a, const Attribute &b);
Attribute operator/(fftw_complex a, const Attribute &b);
Attribute operator/(Vector2 a, const Attribute &b);
Attribute operator/(Vector3 a, const Attribute &b);


inline ostream &operator << (ostream &out_file, const Attribute& thing)
{
  out_file<<"This is an attribute"<<endl;
  return out_file;
}

inline ostream &operator << (ostream &out_file, const BV1& thing)
{
  out_file<<thing.value;
  return out_file;
}







#endif
