/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// The Convolution Kernel -- assumed to square or cubic, and have 2*n+1 elements.
class ConvKernel {
 public:
  ConvKernel();
  ConvKernel(Attribute& a, int nxmax);
  double get_element(int i,int j=0,int k=0);
  int get_nx();
  int get_ndims();
  int set_nx(int nx);
  int set_ndims(int ndims);
  int set_element(double val, int i,int j=0,int k=0);
 private:
  int ndims;
  int nx;
  vector<double> elements;
};

 



/*! Compute the density of a set of Attributes, assumed to be
coordinates.  This version creates a 3-d grid from 3 position
Attributes, and applies a weighting equal to the "mass" attribute. The
returned value is the density field. */
Attribute compute_density(Attribute& x,Attribute& y, Attribute& z, Attribute& mass,
                          vector <int>& nx,vector<double>& ,
                          vector<double>&,int Smoothing=0);
/*! A wrapper for the general compute_density.  In this case, only two
  dimensions are used, and the bounds of the grid are given
  explicitly. */


//!This is the right format to compute the density field.  Note that the grid is returned as a pointer
Grid* compute_density(List* listptr,string xname,string yname, string zname, string massname,
		      string densname,vector<int> nx,
		      vector<double> ll, vector<double> ur);
Grid* compute_density(List* listptr,string xname, string yname, string zname, string massname,
		      string densname, vector<int> nx);
Grid* compute_density(List* listptr,string xname, string yname, string zname, string massname,
		      string densname);

//!2-d with weighting or 3-d without
Grid* compute_density(List* listptr,string xname,string yname, string name3, string densname,
		      vector<int> nx,vector<double> ll,vector<double> ur,int ndims=3);

Grid* compute_density(List* listptr,string xname, string yname, string name3, string densname,
		      vector<int> nx,int ndims=3);
Grid* compute_density(List* listptr, string xname, string yname, string name3, string densname,
		      int ndims=3);

//!1-d with weighting or 2-d without
Grid* compute_density(List* listptr,string xname, string name2,string densname,
		      vector<int> nx,vector<double> ll,
		      vector<double> ur,int ndims=2);
Grid* compute_density(List* listptr,string xname, string name2, string densname, 
		      vector<int> nx,int ndims=2);
Grid* compute_density(List* listptr,string xname, string name2, string densname,
		      int ndims=2);

//! Here we take an existing density field, and make sure to copy the attributes (3d+weighting)
Attribute* compute_density(List* listptr, Grid* gridptr,string xname, string yname, string zname, 
			   string massname, string densname);

//! Here we take an existing density field, and make sure to copy the attributes (2d+weighting or 3d)
Attribute* compute_density(List* listptr,Grid* gridptr,string xname, string yname, string name3, 
			   string densname);

//! Here we take an existing density field, and make sure to copy the attributes (1d+weighting or 2d)
Attribute* compute_density(List* listptr, Grid* gridptr, string xname, string name2, string densname);


// Several wrapped versions for different dimensionality
Attribute compute_density(Attribute& x, Attribute& y, Attribute& z, Attribute& m, vector<int> &nx);
Attribute compute_density(Attribute& x, Attribute& y, Attribute& val3, vector<int> &nx);
Attribute compute_density(Attribute& x, Attribute& val2, vector<int> &nx);
Attribute compute_density(Attribute& x, vector<int> &nx);

//Include the limits as well
Attribute compute_density(Attribute& x, Attribute& y, Attribute& val3, vector<int> &nx, vector<double>& ll,vector<double>& ur);
Attribute compute_density(Attribute& x, Attribute& val2, vector<int> &nx,vector<double>&ll, vector<double>& ur);

//!1 or 2-dim with cube grid
Attribute compute_density(Attribute& x, Attribute& val2, int nx=10,int ndims=2);
//!2 or 3-dim with cube grid
Attribute compute_density(Attribute& x, Attribute& y, Attribute& val3, int nx=10,int ndims=3);
//!3 dim with cube grid
Attribute compute_density(Attribute& x, Attribute& y, Attribute& z, Attribute& m, int nx=10,int ndims=3);
//! 
Attribute compute_density(Attribute& x, int nx=10);



Attribute convolve(Attribute&,ConvKernel);
Attribute& convolve(vector<double> f,ConvKernel c,vector<int> nx, int cot=0);
string equatorial_to_xyz(List* mylist,string ra="",string dec="",string r="",string x="x_conv",
		      string y="y_conv",string z="z_conv");
string xyz_to_equatorial (List* listptr, string x="", string y="", string z="", string ra="ra_conv", string dec="dec_conv", string r="r_conv");
string polar_to_xy (List* listptr, string r="", string theta="", string x="x_conv", string y="y_conv");
string xy_to_polar (List* listptr, string x="", string y="", string r="r_conv", string theta="theta_conv");
string spherical_to_xyz (List* listptr, string rho="", string theta="", string phi="", string x="x_conv", string y="y_conv", string z="z_conv");
string xyz_to_spherical (List* listptr, string x="", string y="", string z="", string rho="rho_conv", string theta="theta_conv", string phi="phi_conv");
string cylindrical_to_xyz (List* listptr, string r="", string theta="", string zIn="", string x="x_conv", string y="y_conv", string z="z_conv");
string xyz_to_cylindrical (List* listptr, string x="", string y="", string zIn="", string r="r_conv", string theta="theta_conv", string z="z_conv");

//! Convert from equatorial (spherical) to cartesian coordinates   This version should be deprecated
string equatorial_to_xyz(Attribute* ra, Attribute* theta, Attribute* r,Attribute* x,Attribute* y,Attribute* z);
//!
string xyz_to_equatorial (Attribute* x, Attribute* y, Attribute* z, Attribute* raOut, Attribute* decOut, Attribute* rOut);
string polar_to_xy (Attribute* r, Attribute* theta, Attribute* xOut, Attribute *yOut);
string xy_to_polar (Attribute* x, Attribute* y, Attribute* rOut, Attribute *thetaOut);
string spherical_to_xyz (Attribute* rho, Attribute* theta, Attribute* phi, Attribute* xOut, Attribute* yOut, Attribute* zOut);
string xyz_to_spherical (Attribute* x, Attribute* y, Attribute* z, Attribute* rhoOut, Attribute* thetaOut, Attribute* phiOut);
string cylindrical_to_xyz (Attribute* r, Attribute* theta, Attribute* z, Attribute* xOut, Attribute* yOut, Attribute* zOut);
string xyz_to_cylindrical (Attribute* x, Attribute* y, Attribute* z, Attribute* rOut, Attribute* thetaOut, Attribute* zOut);


#ifdef BASIN_HAVE_FFTW
Attribute& fft(Attribute& attr_in);
Attribute& fft_c2r(Attribute& attr_in);
Attribute& fft_r2c(Attribute& attr_in);
#endif
